"""
NimbleBox LMAO is our general purpose observability tool for any kind of computation you might have.
"""

# drift detection and all
# run.log_dataset(
#     dataset_name='train',
#     features=X_train,
#     predictions=y_pred_train,
#     actuals=y_train,
# )
# run.log_dataset(
#     dataset_name='test',
#     features=X_test,
#     predictions=y_pred_test,
#     actuals=y_test,
# )

from json import dumps, loads
from typing import Dict, Any, List, Optional, Union
from google.protobuf.timestamp_pb2 import Timestamp

import nbox.utils as U
from nbox.utils import logger
from nbox.auth import secret, AuthConfig

# all the sublime -> hyperloop stuff
from nbox.lmao.lmao_rpc_client import (
  InitRunRequest,
  RunLog,
  Project as ProjectProto,
  Serving as ServingProto,
  ServingLogRequest,
  ServingLogResponse,
  AgentDetails,
)
from nbox.lmao.common import get_lmao_stub, get_record

# functional parts
def get_serving_log(
  project_id: str,
  serving_id: str = "",
  key: str = "",
  after: Timestamp = None,
  before: Timestamp = None,
  limit: int = 100,
) -> ServingLogResponse:
  """
  Get a single log by key from the serving. Logs are stored in the following format:
  `<project_id>/<serving_id>/<key>`

  Args:
    project_id: ID of the project to which the serving belongs.
    serving_id: ID of the serving to which the log belongs.
    key: Key of the log to get.
    after: Return only logs after this timestamp.
    before: Return only logs before this timestamp.
    limit: Maximum number of logs to return.

  Returns:
    The log.
  """
  lmao = get_lmao_stub()
  req = ServingLogRequest(
    workspace_id = secret(AuthConfig.workspace_id),
    project_id = project_id,
    serving_id = serving_id,
    key = key,
    limit = limit,
  )
  if after:
    req.after.CopyFrom(after)
  if before:
    req.before.CopyFrom(before)
  out: ServingLogResponse = lmao.get_serving_log(req)
  return out


# the main class

class LmaoLive():
  def __init__(
    self,
    project_id: str,
    serving_id: str = "",
    metadata: Dict[str, Any] = {},
  ):
    """
    {% CallOut variant="warning" label="This is experimental and not ready for production use, can be removed without notice" /%}
    
    Client module for LMAO Live monitoring pipeline, eventually this will be merged with the LMAO client
    which is used for experiment tracking, into a single combined client.
    """
    self.workspace_id = secret(AuthConfig.workspace_id)
    self.project_id = project_id
    self.serving_id = serving_id
    
    # get the stub
    self.lmao = get_lmao_stub()
    p = ProjectProto(
      workspace_id = self.workspace_id,
      project_id = self.project_id
    )
    print(p)
    self.project = self.lmao.get_project(p)
    if self.project is None:
      raise Exception("Could not connect to LMAO, please check your credentials")
    
    agent_details = AgentDetails(
      workspace_id = self.workspace_id,
      nbx_serving_id = "local",
      nbx_model_id = U.SimplerTimes.get_now_str(),
    )
    self.agent = agent_details

    if serving_id:
      action = "Updating"
      s = ServingProto(
        workspace_id = self.workspace_id,
        project_id = self.project_id,
        serving_id = self.serving_id,
        agent = self.agent,
        update_keys = ["agent",],
      )
      if metadata:
        s.config = dumps(metadata)
        s.update_keys.append("config")
      out = self.lmao.update_serving_status(s)
      if out is None:
        raise Exception(f"Failed to update serving {serving_id}, does this serving exist?")
    else:
      action = "Creating"
      s: ServingProto = self.lmao.init_serving(InitRunRequest(
        workspace_id = self.workspace_id,
        project_id = self.project_id,
        config = dumps(metadata),
        agent_details = self.agent,
      ))

    logger.info(
      f"{action} live tracker\n"
      f" project: {project_id}\n"
      f"      id: {s.serving_id}\n"
      f"    link: {secret(AuthConfig.url)}/workspace/{self.workspace_id}/projects/{project_id}/#Live\n"
    )

    self.serving = s
    self._total_logged_elements  = 0 # total number of elements logged

  def log(self, y: Dict[str, Union[int, float, str]]):
    run_log = RunLog(
      workspace_id = self.workspace_id,
      project_id=self.project_id,
      serving_id = self.serving.serving_id,
      log_type = RunLog.LogType.USER
    )
    for k,v in y.items():
      # TODO: @yashbonde replace Record with RecordColumn
      record = get_record(k, v)
      run_log.data.append(record)

    ack = self.lmao.on_serving_log(run_log)
    if not ack.success:
      logger.error(f"  >> Server Error\n{ack.message}")
      raise Exception("Server Error")

    self._total_logged_elements += 1


# helper class


"""
This code is used to manage the authentication of the entire ``nbox`` package. For authentication
it will create a ``.nbx`` in the user's home directory (``~/.nbx``, in case of linux) and store
a file called ``secrets.json``. This folder will also contain more information and files that
are used elsewhere as well ex. files generated when takling to any instance.

Users can chose to set configs at a global level for convenience. Here's a list of configs that can be set:

global.workspace_id = '' # this will set the default workspace id for all commands, users can override this by
  passing the ``--workspace-id`` flag
"""
import os
import json
import requests
import webbrowser
from typing import Dict
from getpass import getpass
from functools import lru_cache

import nbox.utils as U
from nbox.utils import join, logger, lo


class AuthConfig():
  workspace_id = "config.global.workspace_id"
  workspace_name = "config.global.workspace_name"
  _workspace_id = "workspace_id"
  email = "email"
  cache = "cache"
  username = "username"
  access_token = "access_token"
  url = "nbx_url"

  # things for the pod
  nbx_pod_run = "run"
  nbx_pod_deploy = "deploy"

  def items():
    return [AuthConfig.workspace_id, AuthConfig.workspace_name, AuthConfig.cache]

# kept for legacy reasons, remove it when this code (see git blame) is > 4 months old
ConfigString = AuthConfig


class JobDetails(object):
  job_id: str
  run_id: str

# class DeployDetails(object):

class NBXClient:
  def __init__(self, nbx_url = "https://app.nimblebox.ai"):
    """
    Single source for all the secrets file. Can persist across multiple processes.
    """
    os.makedirs(U.env.NBOX_HOME_DIR(), exist_ok=True)
    self.fp = join(U.env.NBOX_HOME_DIR(), "secrets.json")

    access_token = U.env.NBOX_ACCESS_TOKEN("")

    # if this is the first time starting this then get things from the nbx-hq
    if not os.path.exists(self.fp):
      if not access_token:
        logger.info(f"Ensure that you put the email ID you have signed up with!")
        _secrets_url = f"{nbx_url}/secrets"
        logger.info(f"Opening: {_secrets_url}")
        webbrowser.open(_secrets_url)
        access_token = getpass("Access Token: ")
      
      sess = requests.Session()
      sess.headers = {"Authorization": f"Bearer {access_token}"}
      
      # Once we have the access token, we can get the secrets
      _u = f"{nbx_url}/api/v1"
      r = sess.get(_u+"/user/account_details")
      r.raise_for_status()
      try:
        data = r.json()["data"]
        username = data["username"]
        email = data["email"]
      except Exception as e:
        logger.error(f"Could not get the username and email from the response")
        logger.error(f"This should not have happened, please contact NimbleBox support.")
        raise e

      workspace_id = input("Set the default workspace ID (blank means personal): ").strip()
      workspace_details = {"workspace_name": ""}
      
      if workspace_id:
        # fetch some more data about the workspace
        r = sess.get(_u+"/workspace")
        r.raise_for_status()
        try:
          data = r.json()["data"]
          workspace_details = list(filter(lambda x: x["workspace_id"] == workspace_id, data))
          if len(workspace_details) == 0:
            logger.error(f"Could not find the workspace ID: {workspace_id}. Please check the workspace ID and try again.")
            raise Exception("Invalid workspace ID")
          workspace_details = workspace_details[0]
        except Exception as e:
          logger.error(f"Could not get the workspace details from the response")
          logger.error(f"This should not have happened, please contact NimbleBox support.")
          raise e
      else:
        workspace_id = "personal"
      logger.info(f"Setting the default workspace ID to: {workspace_id}")

      # create the objects
      self.secrets = {
        AuthConfig.email: email,
        AuthConfig.access_token: access_token,
        AuthConfig.url: nbx_url,
        AuthConfig.username: username,

        # config values that can be set by the user for convenience
        AuthConfig.workspace_id: workspace_id,
        AuthConfig.workspace_name: workspace_details["workspace_name"],

        # now cache the information about this workspace
        AuthConfig.cache: {workspace_id: {
          "workspace_name": workspace_details["workspace_name"],
          "workspace_id": workspace_id,
        }},
      }
      # for k,v in self.secrets.items():
      #   print(type(k), k, "::", type(v), v)
      with open(self.fp, "w") as f:
        f.write(repr(self))
      os.makedirs(U.join(U.env.NBOX_HOME_DIR(), ".cache"), exist_ok=True)
      os.makedirs(U.join(U.env.NBOX_HOME_DIR(), "relics"), exist_ok=True)
      logger.info("Successfully created secrets!")
    else:
      with open(self.fp, "r") as f:
        self.secrets = json.load(f)
      logger.debug("Successfully loaded secrets!")

  def __repr__(self):
    return json.dumps(self.secrets, indent=2)

  def get(self, item, default=None, reload: bool = False):
    """
    Get the value of the item from the secrets file.

    Args:
      item (str): The item to get the value for
      default (any): The default value to return if the item is not found
      reload (bool): If True, reload the secrets file before getting the value

    Returns:
      any: The value of the item or the default value if the item is not found
    """
    if reload:
      with open(self.fp, "r") as f:
        self.secrets = json.load(f)
    return self.secrets.get(item, default)

  def put(self, item, value = None, persist: bool = False):
    """
    Put the value of the item in the secrets file. If no `value` or `persist` is specified,
    this is a no-op.

    Args:
      item (str): The item to put the value for
      value (any): The value to put
      persist (bool): If True, persist the secrets file after putting the value
    """
    if value:
      self.secrets[item] = value
    if persist:
      with open(self.fp, "w") as f:
        f.write(repr(self))

  def __call__(self, item, default=None, reload: bool = False):
    return self.get(item, default, reload)

  @property
  def workspace_id(self) -> str:
    return self.get(AuthConfig.workspace_id) or self.get(AuthConfig._workspace_id)

  @property
  def nbx_url(self) -> str:
    return self.get(AuthConfig.url)

  @property
  def access_token(self) -> str:
    return self.get(AuthConfig.access_token)
  
  @property
  def username(self) -> str:
    return self.get(AuthConfig.username)

  def get_agent_details(self) -> Dict[str, str]:
    if ConfigString.nbx_pod_run in self.secrets:
      run_data = self.secrets[ConfigString.nbx_pod_run]
      jd = JobDetails()
      jd.job_id = run_data.get("job_id", None)
      jd.run_id = run_data.get("token", None)
      return jd
    # elif ConfigString.nbx_pod_deploy in self.secrets:
    #   return self.secrets[ConfigString.nbx_pod_deploy]
    return {}


def init_secret():
  """
  Initialize the secret object. This is a singleton object that can be used across multiple processes.
  """
  # add any logic here for creating secrets
  if not U.env.NBOX_NO_AUTH(False):
    secret = NBXClient()
    logger.info(lo(
      f"workspace details",
      workspace_id = secret.workspace_id,
      workspace_name =  AuthConfig.workspace_name,
      token_present = len(secret.access_token) > 0,
      nbx_url = secret.nbx_url,
    ))
    return secret
  else:
    logger.info(f"Skipping authentication as NBOX_NO_AUTH is set to True")
  return None

secret = init_secret()


@lru_cache()
def auth_info_pb():
  """
  Get the auth token for the current user.
  """
  from nbox.hyperloop.common.common_pb2 import NBXAuthInfo

  return NBXAuthInfo(
    username = secret(AuthConfig.username),
    workspace_id = secret(AuthConfig.workspace_id) or secret(AuthConfig._workspace_id),
    access_token = secret(AuthConfig.access_token),
  )

def inside_pod():
  return secret(AuthConfig.nbx_pod_run, False) or secret(AuthConfig.nbx_pod_deploy, False)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: common/common.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13\x63ommon/common.proto\x12\x06\x63ommon\"\xd3\x01\n\x04\x43ode\x12\x0e\n\x06s3_url\x18\x01 \x01(\t\x12\x0c\n\x04size\x18\x02 \x01(\x05\x12\x1f\n\x04type\x18\x03 \x01(\x0e\x32\x11.common.Code.Type\x12)\n\x07s3_meta\x18\x04 \x03(\x0b\x32\x18.common.Code.S3MetaEntry\x1a-\n\x0bS3MetaEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"2\n\x04Type\x12\x0b\n\x07NOT_SET\x10\x00\x12\x08\n\x04NBOX\x10\x01\x12\x07\n\x03ZIP\x10\x02\x12\n\n\x06\x42INARY\x10\x03\"\x8a\x01\n\x0bNBXAuthInfo\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x14\n\x0cworkspace_id\x18\x02 \x01(\t\x12\x14\n\x0c\x61\x63\x63\x65ss_token\x18\x03 \x01(\t\x12\x15\n\rrefresh_token\x18\x04 \x01(\t\x12\x12\n\naccess_key\x18\x05 \x01(\t\x12\x12\n\nproject_id\x18\x06 \x01(\t\"S\n\x0fIntegrationData\x12\x15\n\rconnection_id\x18\x01 \x01(\t\x12\x17\n\x0f\x63onnection_name\x18\x02 \x01(\t\x12\x10\n\x08infra_id\x18\x03 \x01(\t\"\x80\x01\n\x08Resource\x12\x0b\n\x03\x63pu\x18\x01 \x01(\t\x12\x0e\n\x06memory\x18\x02 \x01(\t\x12\x0b\n\x03gpu\x18\x03 \x01(\t\x12\x11\n\tgpu_count\x18\x04 \x01(\t\x12\x11\n\tdisk_size\x18\x05 \x01(\t\x12\x0f\n\x07timeout\x18\x06 \x01(\x03\x12\x13\n\x0bmax_retries\x18\x07 \x01(\x05\"N\n\x03Log\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0b\n\x03idx\x18\x02 \x01(\x05\x12\x0b\n\x03log\x18\x03 \x01(\t\x12!\n\x08k8s_info\x18\x04 \x01(\x0b\x32\x0f.common.K8sInfo\"[\n\x07K8sInfo\x12\x0c\n\x04pods\x18\x01 \x03(\t\x12\x12\n\ndeployment\x18\x02 \x01(\t\x12\x0f\n\x07service\x18\x03 \x01(\t\x12\x0b\n\x03hpa\x18\x04 \x01(\t\x12\x10\n\x08\x63ron_job\x18\x05 \x01(\tBAZ?github.com/NimbleBoxAI/jobs-architecture/common_pb/v3;common_pbb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'common.common_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z?github.com/NimbleBoxAI/jobs-architecture/common_pb/v3;common_pb'
  _CODE_S3METAENTRY._options = None
  _CODE_S3METAENTRY._serialized_options = b'8\001'
  _CODE._serialized_start=32
  _CODE._serialized_end=243
  _CODE_S3METAENTRY._serialized_start=146
  _CODE_S3METAENTRY._serialized_end=191
  _CODE_TYPE._serialized_start=193
  _CODE_TYPE._serialized_end=243
  _NBXAUTHINFO._serialized_start=246
  _NBXAUTHINFO._serialized_end=384
  _INTEGRATIONDATA._serialized_start=386
  _INTEGRATIONDATA._serialized_end=469
  _RESOURCE._serialized_start=472
  _RESOURCE._serialized_end=600
  _LOG._serialized_start=602
  _LOG._serialized_end=680
  _K8SINFO._serialized_start=682
  _K8SINFO._serialized_end=773
# @@protoc_insertion_point(module_scope)



from nbox.plugins.base import experimental_warning

experimental_warning("tensorboard")

def load_tensorboard_to_exp():
  """Experimental feature to load your tensorboard logs to nbox LMAO"""
  raise NotImplementedError("This feature is not yet implemented")



# In case of nbox which handles all kinds of weird paths, initialisation is important.
# We are defining init.py that starts the loading sequence

from nbox.utils import logger
from nbox.subway import Sub30
from nbox.init import nbox_grpc_stub, nbox_session, nbox_ws_v1
from nbox.operator import Operator, operator
from nbox.jobs import Job, Serve, Schedule
from nbox.instance import Instance
from nbox.relics import Relics
from nbox.lmao import Lmao, LmaoLive
from nbox.network import zip_to_nbox_folder
from nbox.version import __version__
from nbox.hyperloop.common.common_pb2 import Resource
from nbox.nbxlib.logger import lo
from nbox.projects import Project
from nbox.auth import secret, AuthConfig, auth_info_pb, inside_pod


import re

from nbox.hyperloop.common.common_pb2 import Resource
from nbox.nbxlib.cloud_machines import AWS, GCP, MSAZ

class PresetMachineConfig(object):
  """Generic dataclass for machine configurations"""
  def __init__(
    self,
    cpu: str = "1000m",
    memory: str = "1Gi",
    gpu: str = "0",
    gpu_name: str = "none",
    disk_size: str = "10Gi",
    metadata: dict = {}
  ) -> None:
    self.cpu = cpu
    self.memory = memory
    self.gpu = gpu
    self.gpu_name = gpu_name
    self.disk_size = disk_size
    self.metadata = metadata

  @property
  def resource(self) -> Resource:
    """Resource object as needed by the NBX Jobs + Deploy Pods"""
    return Resource(
      cpu = self.cpu,
      memory = self.memory,
      gpu_count = self.gpu,
      gpu = self.gpu_name,
      disk_size = self.disk_size,
      max_retries = 2,
      timeout = 120_000,
    )

  def __repr__(self) -> str:
    return str(self.resource)


def get_resource_by_name(name: str) -> Resource:
  if not re.match(r"^[\w]+\.[\w]+$", name):
    raise ValueError(f"Invalid resource name: {name}")
  
  class_name, machine = name.lower().split(".")
  if class_name == "aws":
    cfg = getattr(AWS, machine, None)
  elif class_name == "gcp":
    cfg = getattr(GCP, machine, None)
  elif class_name == "msaz":
    cfg = getattr(MSAZ, machine, None)
  else:
    raise ValueError(f"Invalid resource name: {name}, cloud should be one of: aws, gcp, msaz")
  
  if cfg is None:
    raise ValueError(f"Invalid resource name: {name}")
  return cfg.resource



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rsock.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='rsock.proto',
  package='rsock',
  syntax='proto3',
  serialized_options=b'ZBgithub.com/NimbleBoxAI/nimblebox-rsock-server/v1/rsock_pb;rsock_pb',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0brsock.proto\x12\x05rsock\"U\n\x10HandshakeRequest\x12\x0b\n\x03jwt\x18\x01 \x01(\t\x12\x13\n\x0binstanceKey\x18\x02 \x01(\t\x12\x11\n\tlaunchURL\x18\x03 \x01(\t\x12\x0c\n\x04port\x18\x04 \x01(\t\"\"\n\x11HandshakeResponse\x12\r\n\x05token\x18\x01 \x01(\t\")\n\nDataPacket\x12\r\n\x05token\x18\x01 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x32{\n\x05RSock\x12>\n\tHandshake\x12\x17.rsock.HandshakeRequest\x1a\x18.rsock.HandshakeResponse\x12\x32\n\x06Tunnel\x12\x11.rsock.DataPacket\x1a\x11.rsock.DataPacket(\x01\x30\x01\x42\x44ZBgithub.com/NimbleBoxAI/nimblebox-rsock-server/v1/rsock_pb;rsock_pbb\x06proto3'
)




_HANDSHAKEREQUEST = _descriptor.Descriptor(
  name='HandshakeRequest',
  full_name='rsock.HandshakeRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='jwt', full_name='rsock.HandshakeRequest.jwt', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='instanceKey', full_name='rsock.HandshakeRequest.instanceKey', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='launchURL', full_name='rsock.HandshakeRequest.launchURL', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='port', full_name='rsock.HandshakeRequest.port', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=107,
)


_HANDSHAKERESPONSE = _descriptor.Descriptor(
  name='HandshakeResponse',
  full_name='rsock.HandshakeResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='token', full_name='rsock.HandshakeResponse.token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=109,
  serialized_end=143,
)


_DATAPACKET = _descriptor.Descriptor(
  name='DataPacket',
  full_name='rsock.DataPacket',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='token', full_name='rsock.DataPacket.token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='rsock.DataPacket.data', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=145,
  serialized_end=186,
)

DESCRIPTOR.message_types_by_name['HandshakeRequest'] = _HANDSHAKEREQUEST
DESCRIPTOR.message_types_by_name['HandshakeResponse'] = _HANDSHAKERESPONSE
DESCRIPTOR.message_types_by_name['DataPacket'] = _DATAPACKET
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

HandshakeRequest = _reflection.GeneratedProtocolMessageType('HandshakeRequest', (_message.Message,), {
  'DESCRIPTOR' : _HANDSHAKEREQUEST,
  '__module__' : 'rsock_pb2'
  # @@protoc_insertion_point(class_scope:rsock.HandshakeRequest)
  })
_sym_db.RegisterMessage(HandshakeRequest)

HandshakeResponse = _reflection.GeneratedProtocolMessageType('HandshakeResponse', (_message.Message,), {
  'DESCRIPTOR' : _HANDSHAKERESPONSE,
  '__module__' : 'rsock_pb2'
  # @@protoc_insertion_point(class_scope:rsock.HandshakeResponse)
  })
_sym_db.RegisterMessage(HandshakeResponse)

DataPacket = _reflection.GeneratedProtocolMessageType('DataPacket', (_message.Message,), {
  'DESCRIPTOR' : _DATAPACKET,
  '__module__' : 'rsock_pb2'
  # @@protoc_insertion_point(class_scope:rsock.DataPacket)
  })
_sym_db.RegisterMessage(DataPacket)


DESCRIPTOR._options = None

_RSOCK = _descriptor.ServiceDescriptor(
  name='RSock',
  full_name='rsock.RSock',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=188,
  serialized_end=311,
  methods=[
  _descriptor.MethodDescriptor(
    name='Handshake',
    full_name='rsock.RSock.Handshake',
    index=0,
    containing_service=None,
    input_type=_HANDSHAKEREQUEST,
    output_type=_HANDSHAKERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Tunnel',
    full_name='rsock.RSock.Tunnel',
    index=1,
    containing_service=None,
    input_type=_DATAPACKET,
    output_type=_DATAPACKET,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_RSOCK)

DESCRIPTOR.services_by_name['RSock'] = _RSOCK

# @@protoc_insertion_point(module_scope)



from uuid import uuid4
import json

from nbox import logger, lo, Project
from nbox.utils import SimplerTimes
from nbox.plugins.base import import_error

try:
  from fastapi import FastAPI, Response, Request
  from fastapi.middleware import Middleware
except ImportError:
  raise import_error("fastapi")

# so we need a mixture of both the middleware and the decorator because the decorator can modify the response
# but the middleware can do things after response is sent

async def _decorator_middleware(request: Request, call_next):
  response = await call_next(request)
  scope = request.scope
  trace_id = scope.get("nbx_trace_id", "")
  body = b'{"id":"'+ trace_id.encode("utf-8") + b'","out":'
  async for chunk in response.body_iterator:
    body += chunk
  body += b"}"
  body = body.decode("utf-8")
  response.headers["content-length"] = str(len(body))
  response.headers["content-type"] = "application/json"
  return Response(
    content=body,
    status_code=response.status_code,
    headers=response.headers,
    media_type=response.media_type,
    background=response.background,
  )

class LmaoAsgiMiddleware(Middleware):
  def __init__(self, app, tracker):
    self.app = app
    self.tracker = tracker

  # TODO: @yashbonde add a buffer uploading mechanism for the logs once gRPC is ready

  async def __call__(self, scope, receive, send):
    _to_log = scope['type'] == 'http' and scope["path"] not in ["/", "/metadata"]
    if _to_log:
      trace_id = str(uuid4())
      scope["nbx_trace_id"] = trace_id
      st = SimplerTimes.get_now_ns()
    await self.app(scope, receive, send)
    if _to_log:
      data = {
        "id": trace_id,
        "path": scope["path"],
        "latency": SimplerTimes.get_now_ns() - st
      }
      logger.debug(lo("api_log:", data = data))
      if self.tracker:
        self.tracker.log({
          "data": json.dumps(data),
        })


def add_live_tracker(project: Project, app:FastAPI, metadata = {}) -> FastAPI:
  live_tracker = project.get_live_tracker(metadata = metadata)
  app.add_middleware(LmaoAsgiMiddleware, tracker = live_tracker)
  app.middleware("http")(_decorator_middleware)
  return app



"""
These function are meant to make jobs easier and contain the code for what I am calling
`Subway`, which converts any OpenAPI spec into an RPC-like interface. At the end of the
day, each API call is really nothing but a functional call with added steps like
serialisation and networking combined. This kind of added complexity can:

- Make code look really ugly, think try-catches and `r.raise_for_request()`
- Easy to develop because information has to be packaged as different parts for each
  type of function call (GET, POST, ...). The true intent was always to pass it the
  relevant information and forget about underlying details.
- API endpoints are just strings and it's easy to have a typo in the URL. When the
  `Subway`is loaded with the OpenAPI spec, it will disallow incorrect URLs. while
  managing the parameters correctly.
- A more pythonic way of programming where you can use the dot notation.

Based os these ideas there are three types of subways:

- `Subway`: Does not load OpenAPI spec and will blindly call the API, avoid this
- `Sub30`: Built for OpenAPI v3.0.0, this becomes `nbox_ws_v1`
- `SpecSubway`: This is used with the FastAPI's OpenAPI spec, this is used in systems
  that use FastAPI (eg. Compute Server)
"""

import re
import time
import string
import threading
from google.protobuf.message import Message
from requests import Session
from functools import lru_cache
from json import dumps as json_dumps

from nbox.utils import logger

TIMEOUT_CALLS = 60


class Subway():
  """Simple code that allows extending things by `.attr.ing` them"""
  def __init__(self, _url, _session):
    self._url = _url.rstrip('/')
    self._session = _session

  def __repr__(self):
    return f"<Subway ({self._url})>"

  def __getattr__(self, attr: str):
    return Subway(f"{self._url}/{attr}", self._session)

  def u(self, attr: str):
    return getattr(self, attr)

  def __call__(self, method = "get", trailing = "", json = {}, data = None, _verbose = False, **kwargs):
    fn = getattr(self._session, method)
    url = f"{self._url}{trailing}"
    if _verbose:
      logger.debug(f"Calling {url}")
    items = {}
    if json:
      items["json"] = json
    if data:
      items["data"] = data
    r = fn(url, **items, **kwargs)
    if _verbose:
      logger.debug(r.content.decode())
    try:
      r.raise_for_status() # good when server is good
      return r.json()
    except:
      return r.content.decode()


@lru_cache
def filter_templates(paths):
  re_temps = []
  for x in paths:
    if "{" in x:
      temp_str = "^"
      for y in string.Formatter().parse(x):
        temp_str += y[0]
        if y[1]:
          temp_str += "\w+"
      re_temps.append((re.compile(temp_str + "$"), x,))
    else:
      re_temps.append((re.compile("^"+x+"$"), x,))
  return re_temps


class Sub30:
  def __init__(self, _url, _api, _session: Session, _default_key: str = "data", *, prefix = "", rl = None, bar = None):
    """Like Subway but built for Nimblebox Webserver APIs.

    Examples:
      # Create a stub
      >>> ws = Sub30(
      >>>   _url = "https://my-web.site/",
      >>>   _session = nbox_session,
      >>>   _api = loads(fetch("https://my-web.site/openapi.json", True).decode()),
      >>> )

      # Call things as if they are functions
      >>> ws.users()
      >>> ws.users("post", name = "Fiona Apple")

    Args:
      _url (str): the base url of the API.
      _api (dict): OpenAPI json dict.
      _session (requests.Session): Session object to use for requests.
      prefix (str, optional): This is internal, do not use it explicitly.
      last_call_at (float, optional): This is internal, do not use it explicitly.
    """
    self._url = _url.strip("/")
    self._api = _api
    self._session = _session
    self._default_key = _default_key
    self._prefix = prefix

    self._bar = bar or threading.Barrier(2)
    def _rate_limiter(s = 1.0):
      while True:
        self._bar.wait()
        time.sleep(s)

    if rl is None:
      # daemon meaning: it will die when the main thread dies
      # https://stackoverflow.com/questions/190010/daemon-threads-explanation
      rl = threading.Thread(target=_rate_limiter, daemon=True)
      rl.start()
    self._rl = rl # the limiter function

  def _create_keepalive(self):
    """Creates a keepalive thread. Need to define this as a function because sleeping
    can cause it to behave unlike a normal blocking program, so use this carefully."""

    def _keep_alive():
      # This is a hack to keep the session alive.
      # https://github.com/psf/requests/issues/4937
      # another more elegant solution is given here:
      # https://github.com/psf/requests/issues/4937#issuecomment-788899804
      while True:
        try:
          self._session.get(f"{self._url}")
        except Exception as e:
          pass
        
        if self.thread_kill_event.is_set():
          break
        time.sleep(30)

    self.thread = threading.Thread(target = _keep_alive)
    self.thread.start()
    self.thread_kill_event = threading.Event()

  def _stop(self):
    self.thread_kill_event.set()

  def _copy_session(self):
    from copy import deepcopy
    return deepcopy(self._session)

  def __repr__(self):
    return f"<Sub30 ({self._url})>"

  def __getattr__(self, attr):
    return Sub30(f"{self._url}/{attr}", self._api, self._session, prefix=f"{self._prefix}/{attr}", rl = self._rl, bar = self._bar)

  def u(self, attr):
    return self.__getattr__(attr)

  def __call__(self, _method: str = None, **kwargs):
    """Call the API.

    Args:
      _method (str, optional): if only one method is present this will be ignored else "get" will be used. Defaults to None.
    """
    paths = self._api["paths"]
    for v,s in enumerate(self._api["servers"]):
      # {url: "/api/v1"}, {url: "/api/v2"}
      setattr(self, f"v{v}", s["url"])
    params = None
    json = None

    ft = filter_templates(tuple(paths.keys()))
    path = None
    for t, index in ft:
      if re.match(t, self._prefix):
        path = index
        break
    if path == None:
      raise AttributeError(f"No path found for '{self._prefix}'")

    # this is a match
    logger.debug("Matched path: " + path)
    p = paths[index]
    method = tuple(p.keys())[0] if len(p) == 1 else (
      _method if _method != None else "get"
    )
    body = p[method]

    if "parameters" in body and "requestBody" not in body:
      # likely a GET call
      params = kwargs

    if "requestBody" in body:
      # likely a POST call
      content_type = body["requestBody"]["content"]
      if "application/json" in content_type:
        json = {}
        schema_ref = content_type["application/json"]["schema"]
        for p in schema_ref["properties"]:
          if p in schema_ref.get("required", []):
            assert p in kwargs, f"{p} is required but not provided"
            json[p] = kwargs[p]
          else:
            if not p in kwargs and not "default" in schema_ref["properties"][p]:
              json[p] = None
            else:
              json[p] = kwargs[p]
      # elif "plain/text" in content_type:

    # call and return
    path = re.sub(r"\/_", "/", self._url)
    logger.debug(method.upper() + " " + path)
    
    # barrier will create a sync point for the rate limiter, i.e. if this is reached too quick
    # the rate limiter will sleep for the same amount of time, else it would already be waiting
    # for this
    self._bar.wait()
    r = self._session.request(method, path, json = json, params = params)
    try:
      r.raise_for_status()
    except Exception as e:
      logger.error("Could not complete request, here's some info:")
      logger.error("  Status Code: " + str(r.status_code))
      if json:
        logger.error("       Json: " + json_dumps(json, indent=2))
      if params:
        logger.error("     Params: " + json_dumps(params, indent=2))
      # raise e
      cont = r.content.decode()
      if cont.startswith("<html>"):
        cont = ":: HTML page ::"
      logger.error("  " + cont)
      raise

    out = r.json()
    if self._default_key is not None:
      try:
        out = out[self._default_key]
      except KeyError:
        pass
    return out

class SpecSubway():
  def __init__(self, _url, _session, _spec, __name = None):
    """Subway but for fastAPI OpenAPI spec."""
    self._url = _url.rstrip('/')
    self._session = _session
    self._spec = _spec
    self._name = __name

    self._caller = (
      (len(_spec) == 3 and set(_spec) == set(["method", "meta", "src"])) or
      (len(_spec) == 4 and set(_spec) == set(["method", "meta", "src", "response_kwargs_dict"])) or
      "/" in self._spec)

  @classmethod
  def from_openapi(cls, openapi, _url, _session):
    logger.debug("Loading for OpenAPI version latest")
    paths = openapi["paths"]
    spec = openapi["components"]["schemas"]
    
    tree = {}
    for p in tuple(paths.keys()):
      t = tree
      for part in p.split('/')[1:]:
        part = "/" if part == "" else part
        t = t.setdefault(part, {})

    def _dfs(tree, trail = []):
      for t in tree:
        if tree[t] == {}:
          src = "/" + "/".join(trail)
          if t!= "/":
            src = src + "/" if src != "/" else src
            src = src + t
          
          try:
            data = paths[src]
          except:
            src = src + "/"
            data = paths[src]
          method = tuple(data.keys())[0]
          body = data[method]
          dict_ = {"method": method, "meta": None, "src": src}
          if "requestBody" in body:
            schema_ref = body["requestBody"]["content"]["application/json"]["schema"]["$ref"].split("/")[-1]
            _req_body = spec[schema_ref]
            kwargs_dict = list(_req_body["properties"])
            dict_["meta"] = {
              "kwargs_dict": kwargs_dict,
              "required": _req_body.get("required", None)
            }
          if "responses" in body:
            schema = body["responses"]["200"]["content"]["application/json"]["schema"]
            if "$ref" in schema:
              schema_ref = schema["$ref"].split("/")[-1]
              _req_body = spec[schema_ref]
              kwargs_dict = list(_req_body["properties"])
              if dict_["meta"] != None:
                dict_["meta"].update({"response_kwargs_dict": kwargs_dict})
              else:
                dict_["meta"] = {"response_kwargs_dict": kwargs_dict}
          tree[t] = dict_
        else:
          _dfs(tree[t], trail + [t])

    _dfs(tree)
    
    return cls(_url, _session, tree)

  def __repr__(self):
    return f"<SpecSubway ({self._url})>"

  def __getattr__(self, attr) -> 'SpecSubway':
    # https://stackoverflow.com/questions/3278077/difference-between-getattr-vs-getattribute
    if self._caller and len(self._spec) == 1:
      raise AttributeError(f"'.{self._name}' does not have children")
    if attr not in self._spec:
      raise AttributeError(f"'.{attr}' is not a valid function")
    return SpecSubway(f"{self._url}/{attr}", self._session, self._spec[attr], attr)

  def u(self, attr):
    return self.__getattr__(attr)
  
  def __call__(self, *args, _verbose = False, _parse = False, **kwargs):
    # from pprint import pprint
    # pprint(self._spec)
    if not self._caller:
      raise AttributeError(f"'.{self._name}' is not an endpoint")
    spec = self._spec
    if self._caller and "/" in self._spec:
      spec = self._spec["/"]
    
    data = None
    if spec["meta"] == None:
      assert len(args) == len(kwargs) == 0, "This method does not accept any arguments"
    else:
      spec_meta = spec["meta"]
      if "kwargs_dict" not in spec_meta:
        assert len(args) == len(kwargs) == 0, "This method does not accept any arguments"
      else:
        kwargs_dict = spec["meta"]["kwargs_dict"]
        required = spec["meta"]["required"]
        data = {}
        for i in range(len(args)):
          if required != None:
            data[required[i]] = args[i]
          else:
            data[kwargs_dict[i]] = args[i]
        for key in kwargs:
          if key not in kwargs_dict:
            raise ValueError(f"{key} is not a valid argument")
          data[key] = kwargs[key]
        if required != None:
          for key in required:
            if key not in data:
              raise ValueError(f"{key} is a required argument")

    fn = getattr(self._session, spec["method"])
    url = f"{self._url}"
    if self._caller and "/" in self._spec:
      url += "/"
    # if _verbose:
    logger.debug(f"{spec['method'].upper()} {url}")
    logger.debug(f"-->> {data}")
    r = fn(url, json = data)
    if not r.status_code == 200:
      raise ValueError(r.content.decode())
    
    out = r.json()
    if _parse and self._spec["meta"] != None and "response_kwargs_dict" in self._spec["meta"]:
      out = [out[k] for k in self._spec["meta"]["response_kwargs_dict"]]
      if len(out) == 1:
        return out[0]
    return out



from nbox.utils import logger, lo

def import_error(*packages) -> Exception:
  return ImportError(f"Please install the {', '.join(packages)} packages to use this nbox plugin")

def experimental_warning(name: str):
  logger.warning(f"Seem's like a curious mind is trying to use a new feature: {name}. This feature is experimental and can be unstable")

class PluginCLI:
  def tensorboard(self):
    from nbox.plugins.tensorboard import load_tensorboard_to_exp
    return {
      "load_to_exp": load_tensorboard_to_exp,
    }



"""
Network functions are gateway between NBX-Services.

{% CallOut variant="success" label="If you find yourself using this reach out to NimbleBox support." /%}

But for the curious mind, many of our services work on gRPC and Protobufs. This network.py
manages the quirkyness of our backend and packs multiple steps as one function.
"""

import os
import re
import grpc
import shlex
import jinja2
import fnmatch
import zipfile
import requests
from subprocess import Popen
from typing import Dict, Union
from tempfile import gettempdir
from datetime import datetime, timezone
from google.protobuf.field_mask_pb2 import FieldMask

import nbox.utils as U
from nbox import messages as mpb
from nbox.auth import secret, AuthConfig, auth_info_pb
from nbox.utils import logger, SimplerTimes
from nbox.version import __version__
from nbox.hyperloop.jobs.dag_pb2 import DAG
from nbox.init import nbox_ws_v1, nbox_grpc_stub, nbox_model_service_stub
from nbox.hyperloop.jobs.job_pb2 import  Job as JobProto
from nbox.hyperloop.common.common_pb2 import Resource, Code
from nbox.hyperloop.deploy.serve_pb2 import ModelRequest, Model
from nbox.jobs import Schedule, Serve, Job
from nbox.hyperloop.jobs.nbox_ws_pb2 import JobRequest, UpdateJobRequest
from nbox.nbxlib.operator_spec import OperatorType as OT


#######################################################################################################################
"""
# Serving

Function related to serving of any model.
"""
#######################################################################################################################


def deploy_serving(
  init_folder: str,
  serving_name: str,
  model_name: str,
  serving_id: str = None,
  workspace_id: str = None,
  resource: Resource = None,
  wait_for_deployment: bool = False,
  model_metadata: Dict[str, str] = {},
  exe_jinja_kwargs: Dict[str, str] = {},
  *,
  _only_zip: bool = False,
):
  """Use the NBX-Deploy Infrastructure

  Args:
    init_folder (str): Path to the code
    serving_name (str): Name of the serving
    model_name (str): Name of the model
    serving_id (str, optional): Serving ID. Defaults to None.
    workspace_id (str, optional): Workspace ID. Defaults to None.
    resource (Resource, optional): Resource. Defaults to None.
    wait_for_deployment (bool, optional): Wait for deployment. Defaults to False.
    model_metadata (dict, optional): Model metadata. Defaults to {}.
    exe_jinja_kwargs (dict, optional): Jinja kwargs. Defaults to {}.
  """
  # check if this is a valid folder or not
  if not os.path.exists(init_folder) or not os.path.isdir(init_folder):
    raise ValueError(f"Incorrect project at path: '{init_folder}'! nbox jobs new <name>")
  logger.info(f"Deployment ID (Name): {serving_id} ({serving_name})")

  if wait_for_deployment:
    logger.warning("Wait for deployment to be implemented, please add your own wait loop for now!")

  model_proto = Model(
    serving_group_id = serving_id,
    name = model_name,
    type = Model.ServingType.SERVING_TYPE_NBOX_OP,
    metadata = model_metadata,
    resource = resource
  )
  model_proto_fp = U.join(gettempdir(), "model_proto.msg")
  mpb.write_binary_to_file(model_proto, model_proto_fp)

  # zip init folder
  exe_jinja_kwargs.update({"model_name": model_name,})
  zip_path = zip_to_nbox_folder(
    init_folder = init_folder,
    id = serving_id,
    workspace_id = workspace_id,
    type = OT.SERVING,
    files_to_copy = {model_proto_fp: "model_proto.msg"},
    **exe_jinja_kwargs,
  )
  if _only_zip:
    logger.info(f"Zip file created at: {zip_path}")
    return zip_path
  else:
    return _upload_serving_zip(
      zip_path = zip_path,
      workspace_id = workspace_id,
      serving_id = serving_id,
      model_proto = model_proto,
    )


def _upload_serving_zip(zip_path: str, workspace_id: str, serving_id: str, model_proto: Model):
  model_proto.code.MergeFrom(Code(
    type = Code.Type.ZIP,
    size = int(max(os.stat(zip_path).st_size/(1024*1024), 1)) # MBs
  ))

  # get bucket URL and upload the data
  response: Model = mpb.rpc(
    nbox_model_service_stub.UploadModel,
    ModelRequest(
      model = model_proto,
      auth_info = auth_info_pb()
    ),
    "Could not get upload URL",
    raise_on_error = True
  )
  model_id = response.id
  deployment_id = response.serving_group_id
  logger.debug(f"model_id: {model_id}")
  logger.debug(f"deployment_id: {deployment_id}")

  # upload the file to a S3 -> don't raise for status here
  s3_url = response.code.s3_url
  s3_meta = response.code.s3_meta

  use_curl = False
  if response.code.size > 10:
    logger.warning(
      f"File {zip_path} is larger than 10 MiB ({response.code.size} MiB), this might take a while\n"
      f"Switching to user/agent: cURL for this upload\n"
      f"Protip:\n"
      f"  - if you are constantly uploading large files, take a look at nbox.Relics, that's the right tool for the job"
    )
    use_curl = True

  if use_curl:
    shell_com = f'curl -X POST -F key={s3_meta["key"]} '
    for k,v in s3_meta.items():
      if k == "key":
        continue
      shell_com += f'-F {k}={v} '
    shell_com += f'-F file="@{zip_path}" {s3_url}'
    logger.debug(f"Running shell command: {shell_com}")
    out = Popen(shlex.split(shell_com)).wait()
    if out != 0:
      logger.error(f"Failed to upload model: {out}, please check logs for this")
      return
  else:
    r = requests.post(url=s3_url, data=s3_meta, files={"file": (s3_meta["key"], open(zip_path, "rb"))})
    try:
      r.raise_for_status()
    except:
      logger.error(f"Failed to upload model: {r.content.decode('utf-8')}")
      return

  # model is uploaded successfully, now we need to deploy it
  logger.info(f"Model uploaded successfully, deploying ...")
  response: Model = mpb.rpc(
    nbox_model_service_stub.Deploy,
    ModelRequest(
      model = Model(
        id = model_id,
        serving_group_id = deployment_id,
        resource = model_proto.resource,
        # feature_gates = {"SetRunMetadata": "yoyoyooyyooyyooyoyoyoy-tag"}
      ),
      auth_info = auth_info_pb(),
    ),
    "Could not deploy model",
    raise_on_error=True
  )

  # write out all the commands for this deployment
  # logger.info("API will soon be hosted, here's how you can use it:")
  # _api = f"Operator.from_serving('{serving_id}', $NBX_TOKEN, '{workspace_id}')"
  # _cli = f"python3 -m nbox serve forward --id_or_name '{serving_id}' --workspace_id '{workspace_id}'"
  # _curl = f"curl https://api.nimblebox.ai/{serving_id}/forward"
  _webpage = f"{secret(AuthConfig.url)}/workspace/{workspace_id}/deploy/{serving_id}"
  # logger.info(f" [python] - {_api}")
  # logger.info(f"    [CLI] - {_cli} --token $NBX_TOKEN --args")
  # logger.info(f"   [curl] - {_curl} -H 'NBX-KEY: $NBX_TOKEN' -H 'Content-Type: application/json' -d " + "'{}'")
  logger.info(f"  [page] - {_webpage}")

  return Serve(serving_id = serving_id, model_id = model_id)


#######################################################################################################################
"""
# Jobs

Function related to batch processing of any model.
"""
#######################################################################################################################


def deploy_job(
  init_folder: str,
  job_name: str,
  dag: DAG,
  schedule: Schedule,
  resource: Resource,
  workspace_id: str = None,
  job_id: str = None,
  exe_jinja_kwargs = {},
  *,
  _only_zip: bool = False,
  _unittest: bool = False
) -> None:
  """Upload code for a NBX-Job.

  Args:
    init_folder (str, optional): Name the folder to zip
    job_id_or_name (Union[str, int], optional): Name or ID of the job
    dag (DAG): DAG to upload
    workspace_id (str): Workspace ID to deploy to, if not specified, will use the personal workspace
    schedule (Schedule, optional): If ``None`` will run only once, else will schedule the job
    cache_dir (str, optional): Folder where to put the zipped file, if ``None`` will be ``tempdir``
  Returns:
    Job: Job object
  """
  # check if this is a valid folder or not
  if not os.path.exists(init_folder) or not os.path.isdir(init_folder):
    raise ValueError(f"Incorrect project at path: '{init_folder}'! nbox jobs new <name>")
  if (job_name is None or job_name == "") and job_id == "":
    raise ValueError("Please specify a job name or ID")

  # logger.debug(f"deploy_job:\n  init_folder: {init_folder}\n  name: {job_name}\n  id: {job_id}")

  # job_id, job_name = _get_job_data(name = job_name, id = job_id, workspace_id = workspace_id)
  logger.info(f"Job name: {job_name}")
  logger.info(f"Job ID: {job_id}")

  # intialise the console logger
  URL = secret("nbx_url")
  logger.debug(f"Schedule: {schedule}")
  logger.debug("-" * 30 + " NBX Jobs " + "-" * 30)
  logger.debug(f"Deploying on URL: {URL}")

  # create the proto for this Operator
  job_proto = JobProto(
    id = job_id,
    name = job_name or U.get_random_name(True).split("-")[0],
    created_at = SimplerTimes.get_now_pb(),
    schedule = schedule.get_message() if schedule is not None else None, # JobProto.Schedule(cron = "0 0 24 2 0"),
    dag = dag,
    resource = resource
  )
  job_proto_fp = U.join(gettempdir(), "job_proto.msg")
  mpb.write_binary_to_file(job_proto, job_proto_fp)

  if _unittest:
    return job_proto

  # zip the entire init folder to zip
  zip_path = zip_to_nbox_folder(
    init_folder = init_folder,
    id = job_id,
    workspace_id = workspace_id,
    type = OT.JOB,
    files_to_copy = {job_proto_fp: "job_proto.msg"},
    **exe_jinja_kwargs,
  )
  return _upload_job_zip(zip_path, job_proto,workspace_id)

def _upload_job_zip(zip_path: str, job_proto: JobProto, workspace_id: str):
  # determine if it's a new Job based on GetJob API
  try:
    old_job_proto: JobProto = nbox_grpc_stub.GetJob(
      JobRequest(job = JobProto(id = job_proto.id), auth_info = auth_info_pb())
    )
    new_job = old_job_proto.status in [JobProto.Status.NOT_SET, JobProto.Status.ARCHIVED]
  except grpc.RpcError as e:
    if e.code() == grpc.StatusCode.NOT_FOUND:
      new_job = True
      old_job_proto = JobProto()
    else:
      raise e

  if not new_job:
    # incase an old job exists, we need to update few things with the new information
    logger.debug("Found existing job, checking for update masks")
    paths = []
    mask = mpb.field_mask(old_job_proto, job_proto)
    if old_job_proto.resource.SerializeToString(deterministic = True) != job_proto.resource.SerializeToString(deterministic = True):
      paths.append("resource")
    if old_job_proto.schedule.cron != job_proto.schedule.cron:
      paths.append("schedule.cron")
    logger.debug(f"Updating fields: {paths}")
    nbox_grpc_stub.UpdateJob(
      UpdateJobRequest(job = job_proto, update_mask = FieldMask(paths=paths), auth_info=auth_info_pb()),
    )

  # update the JobProto with file sizes
  job_proto.code.MergeFrom(Code(
    size = max(int(os.stat(zip_path).st_size / (1024 ** 2)), 1), # jobs in MB
    type = Code.Type.ZIP,
  ))

  # UploadJobCode is responsible for uploading the code of the job
  response: JobProto = mpb.rpc(
    nbox_grpc_stub.UploadJobCode,
    JobRequest(job = job_proto, auth_info=auth_info_pb()),
    f"Failed to upload job: {job_proto.id} | {job_proto.name}"
  )
  job_proto.MergeFrom(response)
  s3_url = job_proto.code.s3_url
  s3_meta = job_proto.code.s3_meta
  
  logger.info(f"Uploading model to S3 ... (fs: {response.code.size:0.3f} MB)")
  use_curl = False
  if response.code.size > 10:
    logger.info(
      f"File {zip_path} is larger than 10 MB ({response.code.size} MB), this might take a while\n"
      f"Switching to user/agent: cURL for this upload\n"
      f"Protip:\n"
      f"  - if you are constantly uploading large files, take a look at nbox.Relics, that's the right tool for the job"
    )
    use_curl = True

  if use_curl:
    shell_com = f'curl -X POST -F key={s3_meta["key"]} '
    for k,v in s3_meta.items():
      if k == "key":
        continue
      shell_com += f'-F {k}={v} '
    shell_com += f'-F file="@{zip_path}" {s3_url}'
    logger.debug(f"Running shell command: {shell_com}")
    out = Popen(shlex.split(shell_com)).wait()
    if out != 0:
      logger.error(f"Failed to upload model: {out}, please check logs for this")
      return
  else:
    r = requests.post(url=s3_url, data=s3_meta, files={"file": (s3_meta["key"], open(zip_path, "rb"))})
    try:
      r.raise_for_status()
    except:
      logger.error(f"Failed to upload model: {r.content.decode('utf-8')}")
      return

  job_proto.feature_gates.update({
    "UsePipCaching": "", # some string does not honour value
    "EnableAuthRefresh": ""
  })
  auth_info = auth_info_pb()
  if new_job:
    logger.info("Creating a new job")
    mpb.rpc(
      stub = nbox_grpc_stub.CreateJob,
      message = JobRequest(job = job_proto, auth_info = auth_info),
      err_msg = "Failed to create job"
    )

  if not old_job_proto.feature_gates:
    logger.info("Updating feature gates")
    mpb.rpc(
      stub = nbox_grpc_stub.UpdateJob,
      message = UpdateJobRequest(job = job_proto, update_mask = FieldMask(paths = ["feature_gates"]), auth_info = auth_info),
      err_msg = "Failed to update job",
      raise_on_error = False
    )

  # write out all the commands for this job
  # logger.info("Run is now created, to 'trigger' programatically, use the following commands:")
  # _api = f"nbox.Job(id = '{job_proto.id}', workspace_id='{job_proto.auth_info.workspace_id}').trigger()"
  # _cli = f"python3 -m nbox jobs --job_id {job_proto.id} --workspace_id {workspace_id} trigger"
  # _curl = f"curl -X POST {secret(AuthConfig.url)}/api/v1/workspace/{job_proto.auth_info.workspace_id}/job/{job_proto.id}/trigger"
  # logger.info(f" [python] - {_api}")
  # logger.info(f"    [CLI] - {_cli}")
  # logger.info(f"   [curl] - {_curl} -H 'authorization: Bearer $NBX_TOKEN' -H 'Content-Type: application/json' -d " + "'{}'")

  _webpage = f"{secret(AuthConfig.url)}/workspace/{workspace_id}/jobs/{job_proto.id}"
  logger.info(f"   [page] - {_webpage}")

  # create a Job object and return so CLI can do interesting things
  return Job(job_id = job_proto.id)


#######################################################################################################################
"""
# Common

Function related to both NBX-Serving and NBX-Jobs
"""
#######################################################################################################################

def zip_to_nbox_folder(
  init_folder: str,
  id: str,
  workspace_id: str,
  type: Union[OT.JOB, OT.SERVING],
  files_to_copy: Dict[str, str] = {},
  **jinja_kwargs
):
  """
  This function creates the zip file that is ulitimately uploaded to NBX-Serving or NBX-Jobs.

  Args:
    init_folder (str): The folder that contains the files to be zipped
    id (str): The id of the job or serving
    workspace_id (str): The workspace id of the job or serving
    type (OT_TYPE): The type of the job or serving, comes from nbxlib.operator_spec.OT_TYPE
    files_to_copy (Dict[str, str], optional): A dictionary of source to target filepaths to copy to the zip file. Defaults to {}.
    **jinja_kwargs: Any additional arguments to be passed to the jinja template
  """
  # zip all the files folder
  all_f = U.get_files_in_folder(init_folder, followlinks = False)

  # find a .nboxignore file and ignore items in it
  to_ignore_pat = []
  to_ignore_folder = []
  for f in all_f:
    if f.split("/")[-1] == ".nboxignore":
      with open(f, "r") as _f:
        for pat in _f:
          pat = pat.strip()
          if pat.endswith("/"):
            to_ignore_folder.append(pat)
          else:
            to_ignore_pat.append(pat)
      break

  # print(all_f)

  # print("to_ignore_pat:", to_ignore_pat)
  # print("to_ignore_folder:", to_ignore_folder)

  # two different lists for convinience
  to_remove = []
  for ignore in to_ignore_pat:
    if "*" in ignore:
      x = fnmatch.filter(all_f, ignore)
    else:
      x = [f for f in all_f if f.endswith(ignore)]
    to_remove.extend(x)
  to_remove_folder = []
  for ignore in to_ignore_folder:
    for f in all_f:
      if re.search(ignore, f):
        to_remove_folder.append(f)
  to_remove += to_remove_folder
  all_f = [x for x in all_f if x not in to_remove]
  logger.info(f"Will zip {len(all_f)} files")
  # print(all_f)
  # exit()
  # print(to_remove)
  # exit()

  # zip all the files folder
  zip_path = U.join(gettempdir(), f"nbxjd_{id}@{workspace_id}.nbox")
  logger.info(f"Packing project to '{zip_path}'")
  with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED) as zip_file:
    abspath_init_folder = os.path.abspath(init_folder)
    for f in all_f:
      arcname = f[len(abspath_init_folder)+1:]
      logger.debug(f"Zipping {f} => {arcname}")
      zip_file.write(f, arcname = arcname)

    # create the exe.py file
    exe_jinja_path = U.join(U.folder(__file__), "assets", "exe.jinja")
    exe_path = U.join(gettempdir(), "exe.py")
    logger.debug(f"Writing exe to: {exe_path}")
    with open(exe_jinja_path, "r") as f, open(exe_path, "w") as f2:
      # get a timestamp like this: Monday W34 [UTC 12 April, 2022 - 12:00:00]
      _ct = datetime.now(timezone.utc)
      _day = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"][_ct.weekday()]
      created_time = f"{_day} W{_ct.isocalendar()[1]} [ UTC {_ct.strftime('%d %b, %Y - %H:%M:%S')} ]"

      # fill up the jinja template
      code = jinja2.Template(f.read()).render({
        "created_time": created_time,
        "nbox_version": __version__,
        **jinja_kwargs
      })
      f2.write(code)
    # print(os.stat(exe_path))

    zip_file.write(exe_path, arcname = "exe.py")
    for src, trg in files_to_copy.items():
      zip_file.write(src, arcname = trg)

  return zip_path



"""
Creates a socket tunnel between users localhost to server called RSockServer (Reverse Socket Server) .

```bash
nbx tunnel 8000 --i "nbox-dev"
```

{% CallOut variant="success" label="If you find yourself using this reach out to NimbleBox support." /%}
"""

import os
import sys
import ssl
import socket
import socket
import threading
import subprocess
from time import sleep
from typing import List

import grpc
from nbox.utils import logger as nbx_logger, FileLogger
from nbox import utils as U
from nbox.auth import secret, AuthConfig
from nbox.instance import Instance
from nbox.sub_utils.rsock_pb.rsock_pb2_grpc import RSockStub
from nbox.sub_utils.rsock_pb.rsock_pb2 import HandshakeRequest, HandshakeResponse, DataPacket

class RSockClient:
  # This is a RSockClient. It handels the client socket where client is the user application trying to connect to "client_port"
  # Connects to RSockServer listening on localhost:886.
  # RSockServer recieves instructions as a string and sends back a response.
  # RSockServer requires following steps to setup
  # First,
  #   Authentication:
  #     - Authentication happens by sending
  #       `"AUTH~{AUTH_TOKEN}"`
  #     - AUTH_TOKEN is not defined and is default to 'password'
  #   Seting config:
  #     - you can set config by sending
  #       `"SET_CONFIG~{instance}~{instance_port}"`
  #     - "instance" - Currently is the internal ip of the instance.
  #     - "instance_port" - What port users wants to connect to.
  #   Connect:
  #     - This Starts the main loop which
  #       1. Listen on client_port
  #       2. On connection, 
  #     2. On connection, 
  #       2. On connection, 
  #         a. Send AUTH
  #         b. If AUTH is successful, send SET_CONFIG
  #         c. If SET_CONFIG is successful, send CONNECT
  #         d. If CONNECT is successful, start io_copy
  #   IO_COPY:
  #     - This is the main loop that handles the data transfer between client and server. This is done by creating a new thread for each connection.
  #     - The thread is created by calling the function "io_copy" for each connection that is "server" and "client".
  #     - When a connection is closed, the loop is stopped.

  def __init__(self, connection_id, client_socket, user, subdomain, launch_url, instance_port, file_logger, auth, secure=False):
    """
    Initializes a reverse sockets client.

    Args:
      client_socket: The socket that the client is connected to.
      instance: The instance that the client wants to connect to.
      instance_port: The port that the instance is listening on.
      auth: The authentication token that the client has to provide to connect to the RSockServer.
      secure: Whether or not the client is using SSL.
    """
    self.connection_id = connection_id
    self.client_socket = client_socket
    self.user = user
    self.subdomain = subdomain
    self.launch_url: str = launch_url
    self.instance_port = instance_port
    self.auth = auth
    self.secure = secure

    self.client_auth = False
    self.rsock_thread_running = False
    self.client_thread_running = False
    self.rsock_connection_token = None
    self.logger = file_logger

    self.log('Starting client')
    self.connect_to_rsock_server()
    self.log('Connected to RSockServer')
    self.authenticate()
    # self.log('Authenticated client')
    self.log('Client init complete')


  def __repr__(self):
    return f"RSockClient({self.subdomain} | {self.connection_id})"
  
  def log(self, message, level="INFO"):
    self.logger.info(f"[{self.connection_id}] [{level}] {message}")
  
  def connect_to_rsock_server(self):
    """
    Connects to RSockServer.
    """
    self.log('Connecting to RSockServer', "DEBUG")
    token_cred = grpc.access_token_call_credentials(secret("access_token"))
    ssl_cred = grpc.ssl_channel_credentials()
    creds = grpc.composite_channel_credentials(ssl_cred, token_cred)
    channel = grpc.secure_channel(f'{self.launch_url.replace("https://","").replace(self.subdomain,"rsock")}:443', creds)

    stub = RSockStub(channel)
    self.rsock_stub = stub

    # rsock_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # rsock_socket.connect(('rsock.nimblebox.ai', 886))
    # if self.secure:
    #   self.log('Starting SSL')
    #   certfile = U.join(U.folder(__file__), "pub.crt")
    #   self.rsock_socket = ssl.wrap_socket(rsock_socket, ca_certs=certfile, cert_reqs=ssl.CERT_REQUIRED)
    # else:
    #   self.rsock_socket = rsock_socket

  def authenticate(self):
    """
    Authenticates the client.
    Sends `"AUTH~{AUTH_TOKEN}"` to RSockServer.
    """
    try:
      resp: HandshakeResponse = self.rsock_stub.Handshake(HandshakeRequest(
        jwt = self.auth,
        instanceKey = self.subdomain,
        launchURL = self.launch_url,
        port = str(self.instance_port),
      ))
      self.rsock_connection_token = resp.token
      self.client_auth = True
    except grpc.RpcError as e:
      self.log(f"Authentication failed: {e}", "ERROR")
      exit(1)
    
    

  def connect(self):
    """
    Connects the client to RSockServer.
    Sends `"CONNECT"` to RSockServer.
    """
    if self.client_auth:
      stream_iterator = None
      try:
        self.rsock_thread_running = True
        self.client_thread_running = True
        stream_iterator = self.rsock_stub.Tunnel(self.client_stream())
        # start the io_copy loop
      except grpc.RpcError as e:
        self.log(f"Connection failed: {e}", "ERROR")
        exit(1)

      self.rsock_thread = threading.Thread(target=self.server_stream, args=(stream_iterator, ))
     

      self.rsock_thread.start()
    else:
      self.log('Client authentication failed', "ERROR")
      exit(1)

  def client_stream(self):
    while self.client_thread_running:
      data = self.client_socket.recv(1024)
      if data:
        yield DataPacket(
          token = self.rsock_connection_token,
          data = data,
        )
  
  def server_stream(self,stream):
    try:
      for response in stream:
        self.client_socket.sendall(response.data)
    except grpc.RpcError as e:
      self.log(f"Connection failed: {e}", "ERROR")
      exit(1)
  def stop(self):
    """
    Stops the client.
    """
    self.log('Stopping client')
    self.rsock_thread_running = False
    self.client_thread_running = False
    self.client_socket.close()
    self.log('Client stopped')


class ConnectionManager:
  def __init__(self, user: str, subdomain: str, file_logger: str, auth: str, launch_url:str, notsecure: bool = False):
    """
    Args:
      localport: The port that the client will be listening on.
      user: The user that the client will be connecting as.
      subdomain: The subdomain that the client will be connecting to.
      port: The port that the server will be listening on.
      auth: The build auth token that the client will be using.
      notsecure: Whether or not to use SSL.
    """
    self.user = user
    self.subdomain = subdomain
    self.file_logger = file_logger
    self.auth = auth
    self.launch_url = launch_url
    self.notsecure = notsecure
    self.connection_id = 0

    self.done = False
    self.clients = []
    self.threads = []
    self.connections = {}

  def __repr__(self) -> str:
    return f"ConnectionManager({self.subdomain}: {len(self.threads)} Connections)"

  def add(self, localport, buildport, *, _ssh: bool = True):
    """
    Adds a client to the list of clients.
    """
    nbx_logger.debug(f"Creating a connection: [NBX] {buildport} => {localport} [Local]")
    t = threading.Thread(target = self.start, args = (localport, buildport, _ssh))
    t.start()
    self.threads.append(t)
    self.connections[localport] = buildport

  def start(self, localport, buildport, _ssh: bool = True):
    listen_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    if not _ssh:
      listen_socket.settimeout(1.0)
    listen_socket.bind(('localhost', localport))
    listen_socket.listen(20)

    while not self.done:
      nbx_logger.debug("Waiting for connection on port {}".format(localport))
      try:
        client_socket, _ = listen_socket.accept()
      except socket.timeout:
        break

      nbx_logger.debug('Client connected')
      self.connection_id += 1
      nbx_logger.debug(f'Total clients connected -> {self.connection_id}')

      # create the client
      secure = not self.notsecure
      client = RSockClient(self.connection_id, client_socket, self.user, self.subdomain, self.launch_url, buildport, self.file_logger, self.auth, secure)
      self.clients.append(client)

      # start the client
      client.connect()
    listen_socket.close()

  def quit(self):
    self.done = False
    sleep(2)
    for client in self.clients:
      client.stop()
    for thread in self.threads:
      thread.join()


def port_in_use(port: int) -> bool:
  with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
    return s.connect_ex(('localhost', port)) == 0


def _create_threads(port: int, *apps_to_ports: List[str], i: str, workspace_id: str, _ssh: bool = True) -> ConnectionManager:
  def _sanity():
    if not (sys.platform.startswith("linux") or sys.platform == "darwin"):
      raise Exception(f"Unsupported platform '{sys.platform}', raise issue: https://github.com/NimbleBoxAI/nbox/issues")

    # sanity checks because python fire does not handle empty strings
    if i == "":
      raise Exception("Instance name cannot be empty")
    if workspace_id == "":
      raise Exception("Workspace ID cannot be empty")

  _sanity() # run sanity checks

  apps = {} # <localport-cloudport>
  for ap in apps_to_ports:
    try:
      localport, buildport = ap.split(':')
      apps[int(localport)] = int(buildport)
    except ValueError:
      raise ValueError(f"Incorrect local:build '{ap}', are you passing integers?")
  apps[port] = 2222 # hard code

  ports_used = []
  for k in apps:
    if port_in_use(k):
      ports_used.append(str(k))

  if ports_used:
    raise ValueError(f"Ports {', '.join(ports_used)} are already in use")

  # check if instance is the correct one
  instance = Instance(i, workspace_id = workspace_id)
  instance._unopened_error()

  # create logging for RSock
  folder = U.join(U.env.NBOX_HOME_DIR(), "tunnel_logs")
  os.makedirs(folder, exist_ok=True)
  filepath = U.join(folder, f"tunnel_{instance.project_id}.log") # consistency with IDs instead of names
  file_logger = FileLogger(filepath)
  nbx_logger.info(f"Logging to {filepath}")

  # start the instance with _ssh mode
  instance.start(_ssh = True)

  # conman runs threads internally for multiple connections
  conman = ConnectionManager(
    file_logger = file_logger,
    user = secret("username"), 
    subdomain = instance.open_data.get("url"),
    auth = instance.open_data.get("token"),
    launch_url = instance.open_data.get("launch_url"),
  )
  for localport, buildport in apps.items():
    nbx_logger.debug(f"Creating connection from NBX:{buildport} -> local:{localport}")
    conman.add(localport, buildport, _ssh = _ssh)
  return conman


def tunnel(port: int, *apps_to_ports: List[str], i: str, workspace_id: str = ""):
  """the nbox way to SSH into your instance.

  Usage:
    nbx tunnel 8000 -i "nbox-dev"

  Args:
    port: Local port for terminal
    *apps_to_ports: A tuple of values `buildport:localport`. For example, `jupyter:8888` or `2001:8002`
    i(str): The instance to connect to
  """
  workspace_id: str = workspace_id or secret(AuthConfig.workspace_id)
  connection = _create_threads(port, *apps_to_ports, i = i, workspace_id = workspace_id)

  try:
    # start the ssh connection on terminal    
    # there is more to know about passing shell = True, see:
    # https://medium.com/python-pandemonium/a-trap-of-shell-true-in-the-subprocess-module-6db7fc66cdfd
    # https://stackoverflow.com/questions/3172470/actual-meaning-of-shell-true-in-subprocess
    nbx_logger.info(f"Starting SSH ... for graceful exit press Ctrl+D then Ctrl+C")
    comm = "ssh"
    if U.env.NBOX_SSH_NO_HOST_CHECKING(False):
      comm += " -o StrictHostKeychecking=no"
    comm += f" -p {port} ubuntu@localhost"
    nbx_logger.debug(f"Running command: {comm}")
    subprocess.call(comm, shell=True)
  except KeyboardInterrupt:
    nbx_logger.info("KeyboardInterrupt, closing connections")
    connection.quit()

  connection.quit()
  sys.exit(0) # graceful exit